//===========================================================================
//=
//=  Copyright (C) 2008 Maxim Integrated Products
//=  All rights Reserved. Printed in U.S.A.
//=
//=  Permission is hereby granted, free of charge, to any person obtaining a
//=  copy of this software and associated documentation files (the 
//=  "Software"), to deal in the Software without restriction, including
//=  without limitation the rights to use, copy, modify, merge, publish,
//=  distribute, sublicense, and/or sell copies of the Software, and to
//=  permit persons to whom the Software is furnished to do so, subject to
//=  the following conditions:
//=  
//=  The above copyright notice and this permission notice shall be included
//=  in all copies or substantial portions of the Software source code.
//=  
//=  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
//=  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//=  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//=  IN NO EVENT SHALL MAXIM INTEGRATED PRODUCTS BE LIABLE FOR ANY CLAIM, 
//=  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
//=  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR 
//=  THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//=
//=  Except as contained in this notice, the name of Maxim Integrated Products
//=  shall not be used except as stated in the Maxim Branding Policy.
//=
//=     Description: MAXQ2000 uIP Ethernet Example Application
//=                 
//=        Filename: main.c
//=
//=        Compiler: Rowley CrossWorks C compiler
//=
//=        Hardware: MAXQ2000 Evaluation Kit (Rev B)
//=
//===========================================================================

/* $Id: main.c,v 1.7 2008/02/04 21:21:57 zmetzing Exp $ */

#include <stdio.h>

#include "uip/clock.h"
#include "clock-arch.h"

#include "copybuf.h"
#include "serialio.h"
#include "spi.h"
#include "cprintf.h"
#include "macphy.h"
#include "util.h"
#include "onewire.h"

#include "uip/uip.h"
#include "uip/timer.h"
#include "uip/uip_arp.h"

/* Packet buffer for uip, which is allocated in macphy.c */
extern unsigned char pkt_buf[];

/* Re-transmit buffer length */
unsigned int xmitlen;

#define BUF ((struct uip_eth_hdr *)&uip_buf[0])

void main(void)
{
  unsigned int i;
  uip_ipaddr_t ipaddr;
  unsigned long y;
  struct timer periodic_timer, arp_timer;
  unsigned char ch, outbits, inbits;
  unsigned int sigbits, lastbits, crc;

  /* Build a locally-administered MAC ID just in case we can't read it from the 1-Wire device */
  unsigned char macaddr[] = {0x02, 0x01, 0x02, 0x03, 0x04, 0x05}; 
  struct uip_eth_addr uipmacaddr = {{0x02, 0x01, 0x02, 0x03, 0x04, 0x05}};

  /* Pause to get off of ring */
  for (i = 0; i < 0xffff; i++);

  /* Console */
  serialio_init(0x2f2f); /* 115200 bps for 20MHz HF xtal */

  cprintf(C"\r\nAN4205: MAXQ2000 uIP Example Application\r\n");

  /* Sanity check for buffer mis-alignment along with compiler bug issue */
  cprintf(C"pkt_buf == 0x%04x\r\n", &pkt_buf[0]);
  if ((int)&pkt_buf[0] & 0x0001) {
    cprintf(C" panic: pkt_buf is not word aligned. This build won't work!\r\n");
    while(1);
  }

  /* Start the RTC */
  clock_init();

  /* Program clock rates, etc */
  spi_init();

  /* Retrieve MAC ID from 1-wire DS2502-E48 */
  if (onewire_macid(&macaddr[0]) < 0) {
    cprintf(C"WARNING: [onewire_macid()] 1-wire device not present or error in data\r\n");
    cprintf(C"WARNING: Using default MAC ID\r\n");
  } else {
    for (i = 0; i < 6; i++) {
      uipmacaddr.addr[i] = macaddr[i];
    }
  }
  cprintf(C"Ethernet MAC ID %02X:%02X:%02X:%02X:%02X:%02X\r\n", macaddr[0], macaddr[1],
	  macaddr[2], macaddr[3], macaddr[4], macaddr[5]);
  
  /* Reset the ENC28J60 */
  macphy_reset();

  /* Enable ethernet */
  macphy_init(macaddr);

  /* Begin receiving packets */
  macphy_startrx();

  /* Taken almost verbatim from the uIP example code */
  timer_set(&periodic_timer, CLOCK_SECOND / 8); /* 125uS periodic interrupts */
  timer_set(&arp_timer, CLOCK_SECOND * 10);

  /* Initialize the uIP ARP subsystem */
  uip_arp_init();
  uip_setethaddr(uipmacaddr);

  /* Initialize the uIP stack proper */
  uip_init();

  /* Hardcode the IP address */
  uip_ipaddr(ipaddr, 10, 16, 24, 80);
  uip_sethostaddr(ipaddr);
  cprintf(C"My IP address is %u.%u.%u.%u\r\n", 
	  uip_ipaddr1(ipaddr), uip_ipaddr2(ipaddr), uip_ipaddr3(ipaddr), uip_ipaddr4(ipaddr));
	  
  uip_ipaddr(ipaddr, 255, 255, 252, 0);
  uip_setnetmask(ipaddr);
  cprintf(C"My netmask is %u.%u.%u.%u\r\n", 
	  uip_ipaddr1(ipaddr), uip_ipaddr2(ipaddr), uip_ipaddr3(ipaddr), uip_ipaddr4(ipaddr));

  /* Initialize the application to listen on TCP port 23 (telnet) */
  uip_listen(HTONS(23));

  while(1) {
    
    /* Incoming packet? */
    if (macphy_pktcnt() > 0) {
      macphy_readpkt();
      uip_len = (pkt_buf[3] << 8) + pkt_buf[2];
    } else {
      uip_len = 0;
    }

    /* Outgoing packet? */
    if(uip_len > 0) {
      if(BUF->type == htons(UIP_ETHTYPE_IP)) {
        uip_arp_ipin();
        uip_input();
        /* If the above function invocation resulted in data that
           should be sent out on the network, the global variable
           uip_len is set to a value > 0. */
        if(uip_len > 0) {
          uip_arp_out();
	  macphy_sendpkt();
        }
      } else if(BUF->type == htons(UIP_ETHTYPE_ARP)) {
        uip_arp_arpin();
        /* If the above function invocation resulted in data that
           should be sent out on the network, the global variable
           uip_len is set to a value > 0. */
        if(uip_len > 0) {
	  macphy_sendpkt();
        }
      }

    } else if(timer_expired(&periodic_timer)) {
      timer_reset(&periodic_timer);
      for(i = 0; i < UIP_CONNS; i++) {
        uip_periodic(i);
        /* If the above function invocation resulted in data that
           should be sent out on the network, the global variable
           uip_len is set to a value > 0. */
        if(uip_len > 0) {
          uip_arp_out();
	  macphy_sendpkt();
        }
      }
     
      /* Call the ARP timer function every 10 seconds. */
      if(timer_expired(&arp_timer)) {
        timer_reset(&arp_timer);
        uip_arp_timer();
      }
    }
  }

  while(1);
}

/* Handles the callbacks generated by the uIP stack
 *
 * Input arguments:
 *
 *  none
 *
 * Returns:
 *
 *  none
 *
 */
void uip_callback(void)
{
  unsigned int ptr, ptr0;
  unsigned char tmp, *appdata;

  /* Assign this so as to avoid typecasting */
  appdata = uip_appdata;

  /* New connections to the listening port will be greeted with a banner */
  if (uip_connected()) {
    /* Send a message to the console indicating a new connection */
    cprintf(C"New connection from IP %u.%u.%u.%u remote port %u\r\n", 
	    uip_ipaddr1(uip_conn->ripaddr), uip_ipaddr2(uip_conn->ripaddr),
	    uip_ipaddr3(uip_conn->ripaddr), uip_ipaddr4(uip_conn->ripaddr),
	    uip_conn->rport);

    /* Greet the new connection */
    xmitlen = 15;
    asm_copybuffer(appdata, C"Hello world!\r\n", xmitlen);
    uip_send(appdata, xmitlen);
  }

  /* Closed connections (normal, aborted, or timed out) are handled here */
  if (uip_closed()) {
    /* Send a message to the console indicating a closed connection */
    cprintf(C"Connection closed (IP %u.%u.%u.%u remote port %u)\r\n",
	    uip_ipaddr1(uip_conn->ripaddr), uip_ipaddr2(uip_conn->ripaddr),
	    uip_ipaddr3(uip_conn->ripaddr), uip_ipaddr4(uip_conn->ripaddr),
	    uip_conn->rport);
  }
  if (uip_aborted()) {
    cprintf(C"Connection aborted (IP %u.%u.%u.%u remote port %u)\r\n",
	    uip_ipaddr1(uip_conn->ripaddr), uip_ipaddr2(uip_conn->ripaddr),
	    uip_ipaddr3(uip_conn->ripaddr), uip_ipaddr4(uip_conn->ripaddr),
	    uip_conn->rport);
  }
  if (uip_timedout()) {
    cprintf(C"Connection timed out (IP %u.%u.%u.%u remote port %u)\r\n",
	    uip_ipaddr1(uip_conn->ripaddr), uip_ipaddr2(uip_conn->ripaddr),
	    uip_ipaddr3(uip_conn->ripaddr), uip_ipaddr4(uip_conn->ripaddr),
	    uip_conn->rport);
  }

  /* If we recieve new data from the peer, we'll process it here */
  if (uip_newdata()) {

    cprintf(C"New data on connection. uip_datalen() == %u\r\n", uip_datalen());

    /* Since we're using telnet to connect to this application, we need to discard */
    /* the TELNET option negotiation on first connect */
    if (appdata[0] == 0xff) {
      cprintf(C"Skipping TELNET option negotiation\r\n");

    } else {
      ptr = 0;
      /* Find the last character to reverse */
      while (ptr < uip_datalen()) {
	if ((appdata[ptr] == '\r') ||
	    (appdata[ptr] == '\n')) {
	  /* Back up and done! */
	  ptr--;
	  break;
	} else {
	  ptr++;
	}
      }
      
      /* Reverse the characters */
      ptr0 = ptr;
      while (ptr > (ptr0 - ptr)) {
	/* Save off the mirror character */
	tmp = appdata[ptr0 - ptr];
	/* Reassign it to character at ptr */
	appdata[ptr0 - ptr] = appdata[ptr];
	/* Reassign ptr character to the mirror's old value and decrement */
	appdata[ptr--] = tmp;
      }
      
      /* Transmit it to the peer */
      xmitlen = uip_datalen();
      uip_send(appdata, xmitlen);
    }
  }

  /* uIP provides for a periodic "poll" call, during which an application can do */
  /* periodic tasks and, possibly, generate data to transmit to the peer */
  if (uip_poll()) {
  }

  /* Retransmissions are handled here */
  if (uip_rexmit()) {
    /* Data needed for retransmission is still in appdata. Simply send it again */
    uip_send(appdata, xmitlen);
  }

  return;
}
