//===========================================================================
//=
//=  Copyright (C) 2008 Maxim Integrated Products
//=  All rights Reserved. Printed in U.S.A.
//=
//=  Permission is hereby granted, free of charge, to any person obtaining a
//=  copy of this software and associated documentation files (the 
//=  "Software"), to deal in the Software without restriction, including
//=  without limitation the rights to use, copy, modify, merge, publish,
//=  distribute, sublicense, and/or sell copies of the Software, and to
//=  permit persons to whom the Software is furnished to do so, subject to
//=  the following conditions:
//=  
//=  The above copyright notice and this permission notice shall be included
//=  in all copies or substantial portions of the Software source code.
//=  
//=  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
//=  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//=  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//=  IN NO EVENT SHALL MAXIM INTEGRATED PRODUCTS BE LIABLE FOR ANY CLAIM, 
//=  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
//=  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR 
//=  THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//=
//=  Except as contained in this notice, the name of Maxim Integrated Products
//=  shall not be used except as stated in the Maxim Branding Policy.
//=
//=     Description: MAXQ2000 uIP Ethernet Example Application
//=                 
//=        Filename: onewire.c
//=
//=        Compiler: Rowley CrossWorks C compiler
//=
//=        Hardware: MAXQ2000 Evaluation Kit (Rev B)
//=
//===========================================================================

/* $Id: onewire.c,v 1.2 2008/02/04 21:21:57 zmetzing Exp $ */

#include <maxq2000.h>

#include "onewire.h"
#include "cprintf.h"
#include "util.h"

int onewire_macid(unsigned char *macid)
{
  unsigned int i, ow_crc;
  unsigned char databuffer[16];

  /* Set Clock Divisor for 20MHz */
  OWA = 0x04; /* Clock Divisor */
  OWD = 0x8a;

  /* Attempt to detect one wire device */
  OWA = 0x00; /* Command Register */
  OWD = 0x01; /* Reset! */

  /* Wait here for 1ms */
  for (i = 0; i < 0x4e20; i++);

  /* Check presence-detect */
  OWA = 0x02;
  if ((OWD & 0x02) > 0) {
    /* Nothing found. Indicate error and bail! */
    return -1;
  }

  /* Skip ROM */
  if (onewire_txrx(0xcc, &databuffer[0]) < 0) {
    /* cprintf(C"OW TIMEOUT! (SKIP ROM)\r\n"); */
    return -1;
  }
  /* Read Memory */
  if (onewire_txrx(0xf0, &databuffer[0]) < 0) {
    /* cprintf(C"OW TIMEOUT! (READ MEM)\r\n"); */
    return -1;
  } 
  /* Address Low */
  if (onewire_txrx(0x00, &databuffer[0]) < 0) {
    /* cprintf(C"OW TIMEOUT! (ADDR LOW)\r\n"); */
    return -1;
  } 
  /* Address High */
  if (onewire_txrx(0x00, &databuffer[0]) < 0) {
    /* cprintf(C"OW TIMEOUT! (ADDR HIGH)\r\n"); */
    return -1;
  } 

  /* Command checksum readback */
  if (onewire_txrx(0xff, &databuffer[0]) < 0) {
    /* cprintf(C"OW TIMEOUT! (Command checksum)\r\n"); */
    return -1;
  } 
  if (databuffer[0] != 0x8d) {
    /* Most likely corruption on the 1-wire during TX */
    return -1;
  }
  
  for (i = 0; i < 13; i++) {
    if (onewire_txrx(0xff, &databuffer[i]) < 0) {
      /* cprintf(C"OW TIMEOUT! (DATA)\r\n"); */
      return -1;
    }
  }

  /* Compute CRC over data */
  ow_crc = 0;
  for (i = 0; i < 11; i++) {
    ow_crc = onewire_crc16(ow_crc, databuffer[i]);
  }
  /* Complement this per the iButton standards */
  ow_crc = ~ow_crc;

#if defined(ONEWIRE_DEBUG)
  cprintf(C"Received OW CRC: %02X%02X\r\n", databuffer[12], databuffer[11]);
  cprintf(C"Calculated OW CRC: %04X\r\n", ow_crc);
#endif  

  if (ow_crc != ((databuffer[12] << 8) + databuffer[11])) {
    cprintf(C"OW CRC ERROR\r\n");
    return -1;
  }
    
  macid[0] = databuffer[10];
  macid[1] = databuffer[9];
  macid[2] = databuffer[8];
  macid[3] = databuffer[7];
  macid[4] = databuffer[6];
  macid[5] = databuffer[5];

#if defined(ONEWIRE_DEBUG)
  cprintf(C" -=- One Wire Ethernet ID Response -=-\r\n");
  hexdump(databuffer, 16);
#endif

  return 0;
}

int onewire_txrx(unsigned char tx_data, unsigned char *rx_data)
{
  unsigned int i, tries;

  /* Put data into TX/RX buffer */
  OWA = 0x01;
  OWD = tx_data;

  /* Wait here for 1ms */
  for (i = 0; i < 0x4e20; i++);

  /* Spin on receive buffer full bit */
  OWA = 0x02;
  tries = 0;
  while ((OWD & 0x10) == 0) {
    tries++;
    if (tries > 100) {
      /* Flag error and bail */
      return -1;
    }
  }
  
  /* Get data from buffer */
  OWA = 0x01;
  *rx_data = OWD;

  /* We're good */
  return 0;
}

int onewire_odd_parity(int data)
{
   int i, n = 0;

   for (i = 0; i < 8; ++i)
   {
      n += data & 1;
      data = data >> 1;
   }

   return n & 1;
}

unsigned int onewire_crc16(unsigned int crc, int data)
{
   data = (data ^ (crc & 0xff)) & 0xff;
   crc >>= 8;

   if (onewire_odd_parity(data))
      crc ^= 0xc001;

   data <<= 6;
   crc   ^= data;
   data <<= 1;
   crc   ^= data;

   return crc;
}
